@@ -1,6 +1,6 @@
 ;;; el-search-x.el --- Additional pattern definitions for el-search    -*- lexical-binding: t -*-
 
-;; Copyright (C) 2016, 2017 Free Software Foundation, Inc
+;; Copyright (C) 2016-2018 Free Software Foundation, Inc
 
 ;; Author: Michael Heerdegen <michael_heerdegen@web.de>
 ;; Maintainer: Michael Heerdegen <michael_heerdegen@web.de>
@@ -29,23 +29,31 @@
 ;; This file contains additional definitions of el-search patterns.
 ;; You can just `require' this file, but doing so is not mandatory for
 ;; using el-search.
-;;
-;; TODO:
-;;
-;; - (l (and s) __ (and s)) should match (x y x 1) but
-;;   doesn't.  This seems to happen because `el-search--split'
-;;   considers only _one_ matching split, but we must consider
-;;   _all_.  And (l (and s) __ (pred (eq s))) even errors.
+
 
 
 ;;; Code:
 
 (eval-when-compile
-  (require 'subr-x)
-  (require 'thunk))
+  (require 'subr-x))
+(require 'thunk)
 (require 'el-search)
 
 
+(el-search-defpattern string-lines (pattern)
+  "Matches any string whose line count is matched by PATTERN.
+
+Examples: (string-lines 1) matches one-line strings.
+\(string-lines (pred (>= 5))\) matches strings consisting of not
+more than 5 lines."
+  (let ((string (make-symbol "string")))
+    `(and (string)
+          ,string
+          (let ,pattern
+            (with-temp-buffer
+              (insert ,string)
+              (count-lines (point-min) (point-max)))))))
+
 ;;;; `append and `l'
 
 (defun el-search--split (matcher1 matcher2 list)
@@ -86,30 +94,39 @@ Example: the pattern
 
    (append '(1 2 3) x (app car-safe 7))
 
-matches the list (1 2 3 4 5 6 7 8 9) and binds `x' to (4 5 6)."
+matches the list (1 2 3 4 5 6 7 8 9), binding `x' to (4 5 6)."
   (if (null patterns)
       '(pred null)
     (pcase-let ((`(,pattern . ,more-patterns) patterns))
       (cond
-       ((null more-patterns)  pattern)
+       ((null more-patterns) pattern)
        ((null (cdr more-patterns))
         `(and (pred listp)
               (app ,(apply-partially #'el-search--split
-                                     (el-search--matcher pattern)
-                                     (el-search--matcher (car more-patterns)))
+                                     (el-search-make-matcher pattern)
+                                     (el-search-make-matcher (car more-patterns)))
                    `(,,pattern ,,(car more-patterns)))))
        (t `(append ,pattern (append ,@more-patterns)))))))
 
+(defcustom el-search-lazy-l t
+  "Whether to interpret symbols and strings specially in `l'.
+
+When non-nil, the default, `l' based pattern types interpret
+symbols and strings as special LPATS: a SYMBOL matches any symbol
+S matched by SYMBOL's name interpreted as a regexp, and a STRING
+matches any string matched by the STRING interpreted as a regexp.
+
+When nil, symbols and strings act as standard `pcase' patterns."
+  :group 'el-search :type 'boolean)
+
 (defun el-search--transform-nontrivial-lpat (expr)
-  (pcase expr
-    ((and (pred symbolp) (let symbol-name (symbol-name expr)))
-     `(or (symbol ,symbol-name)
-          `',(symbol  ,symbol-name)
-          `#',(symbol ,symbol-name)))
-    (`',(and (pred symbolp) symbol)
-     `(or ',symbol '',symbol '#',symbol))
-    ((pred stringp) `(string ,expr))
-    (_ expr)))
+  (if el-search-lazy-l
+      (pcase expr
+        ((and (pred symbolp) (let symbol-name (symbol-name expr)))
+         `(symbol ,symbol-name))
+        ((pred stringp) `(string ,expr))
+        (_ expr))
+    expr))
 
 (el-search-defpattern l (&rest lpats)
   "Alternative pattern type for matching lists.
@@ -119,41 +136,45 @@ order.
 The idea is to be able to search for pieces of code (i.e. lists)
 with very brief input by using a specialized syntax.
 
-An LPAT can take the following forms:
+An LPAT can take the following forms (the special interpretation
+of symbols and strings can be turned off by binding or
+customizing `el-search-lazy-l' to nil):
 
 SYMBOL  Matches any symbol S matched by SYMBOL's name interpreted
-        as a regexp.  Matches also 'S and #'S for any such S.
-'SYMBOL Matches SYMBOL, 'SYMBOL and #'SYMBOL (so it's like the above
-        without regexp matching).
+        as a regexp.
+'SYMBOL Matches the SYMBOL.
 STRING  Matches any string matched by STRING interpreted as a
-        regexp
-_       Matches any list element
-__      Matches any number of list elements (including zero)
-^       Matches zero elements, but only at the beginning of a list
-$       Matches zero elements, but only at the end of a list
-PAT     Anything else is interpreted as a normal pcase pattern, and
-        matches one list element matched by it
+        regexp.
+_       Matches any list element.
+__      Matches any number (including zero) of list elements.
+^       Matches zero elements, but only at the beginning of a list.
+        Only allowed as the first of the LPATS.
+$       Matches zero elements, but only at the end of a list.
+        Only allowed as the last of the LPATS.
+PAT     Anything else is interpreted as a standard pattern and
+        matches one list element matched by it.  Note: If
+        matching PAT binds any symbols, occurrences in any
+        following patterns are not turned into equivalence tests;
+        the scope of symbol bindings is limited to the PAT
+        itself.
 
-^ is only valid as the first, $ as the last of the LPATS.
-
-Example: To match defuns that contain \"hl\" in their name and
-have at least one mandatory, but also optional arguments, you
+Example: To match defuns that contain \"hl\" in the defined name
+and have at least one mandatory, but also optional arguments, you
 could use this pattern:
 
     (l ^ 'defun hl (l _ &optional))"
+  ;; We don't allow PATs in `l' to create bindings because to make this
+  ;; work as expected we would need some kind of backtracking
   (declare
    (heuristic-matcher
     (lambda (&rest lpats)
-      (lambda (file-name-or-buffer atom-thunk)
+      (lambda (file-name-or-buffer atoms-thunk)
         (cl-every
          (lambda (lpat)
            (pcase lpat
              ((or '__ '_ '_? '^ '$) t)
-             ((pred symbolp)
-              (funcall (el-search-heuristic-matcher `(symbol ,(symbol-name lpat)))
-                       file-name-or-buffer atom-thunk))
              (_ (funcall (el-search-heuristic-matcher (el-search--transform-nontrivial-lpat lpat))
-                         file-name-or-buffer atom-thunk))))
+                         file-name-or-buffer atoms-thunk))))
          lpats)))))
   (let ((match-start nil) (match-end nil))
     (when (eq (car-safe lpats) '^)
@@ -185,20 +206,50 @@ could use this pattern:
 (defcustom el-search-change-revision-transformer-function nil
   "Transformer function for the REVISION argument of `change' and `changed'.
 
-When specified, this function is called with two arguments: the
-REVISION argument passed to `change' or `changed', and the
-current file name, and the returned value is used instead of
-REVISION.
+When specified, this function is called with two arguments - the
+REVISION argument passed to `change' or `changed' and the current
+file name - and the return value is used as REVISION argument for
+these patterns.
 
 The default value is nil."
   :group 'el-search
   :type '(choice (const :tag "No transformer" nil)
                  (function :tag "User specified function")))
 
+(defalias 'el-search--file-truename-wstm
+  ;; We call `file-truename' very often and it's quite slow
+  (el-search-with-short-term-memory #'file-truename))
+
+(defun el-search--changed-files-in-repo (repo-root-dir &optional commit)
+  "Return a list of files that changed relative to COMMIT.
+COMMIT defaults to HEAD."
+  (cl-callf or commit "HEAD")
+  (let ((default-directory repo-root-dir))
+    (mapcar #'expand-file-name
+            (split-string
+             (shell-command-to-string
+              (format "git diff -z --name-only %s --" (shell-quote-argument commit)))
+             "\0" t))))
+
+(defvar vc-git-diff-switches)
+(defun el-search--file-changed-p (file revision)
+  "Return non-nil when FILE has changed relative to REVISION."
+  (cl-callf el-search--file-truename-wstm file)
+  (when-let ((backend (vc-backend file)))
+    (let ((default-directory (file-name-directory file)))
+      (and
+       (with-temp-buffer
+         (= 1 (vc-call-backend backend 'diff (list file) nil revision (current-buffer))))
+       (with-temp-buffer
+         (= 1 (vc-call-backend backend 'diff (list file) revision nil (current-buffer))))))))
+
 (defun el-search--changes-from-diff-hl (revision)
-  "Return a list of changed regions (as conses of positions) since REVISION.
-Use variable `el-search--cached-changes' for caching."
-  (let ((buffer-file-name (file-truename buffer-file-name))) ;shouldn't be necessary, but it is...
+  "Return the changed regions in the current buffer's file.
+The return value is a list of conses (START . END) of all changes
+relative to REVISION.
+
+Uses variable `el-search--cached-changes' for caching."
+  (let ((buffer-file-name (el-search--file-truename-wstm buffer-file-name))) ;shouldn't be necessary, but it is...
     (if (and (consp el-search--cached-changes)
              (equal (car el-search--cached-changes)
                     (list revision (visited-file-modtime))))
@@ -234,18 +285,29 @@ Use variable `el-search--cached-changes' for caching."
                                                       (let ((default-directory (file-name-directory buffer-file-name)))
                                                         (diff-hl-changes)))))))))))))))
 
-(defun el-search--change-p (posn &optional revision)
+(defun el-search--change-p (posn revision)
   ;; Non-nil when sexp after POSN is part of a change
+  (when (buffer-modified-p)
+    (user-error "Buffer is modified - please save"))
   (save-restriction
     (widen)
     (let ((changes (el-search--changes-from-diff-hl revision))
-          (sexp-end (scan-sexps posn 1)))
-      (while (and changes (< (cdar changes) sexp-end))
+          (sexp-end (el-search--end-of-sexp posn))
+          (atomic? (thunk-delay (el-search--atomic-p
+                                 (save-excursion (goto-char posn)
+                                                 (el-search-read (current-buffer)))))))
+      (while (and changes (or (< (cdar changes) posn)
+                              (and
+                               ;; a string spanning multiple lines is a change even when not all
+                               ;; lines are changed
+                               (< (cdar changes) sexp-end)
+                               (not (thunk-force atomic?)))))
         (pop changes))
-      (and changes
-           (<= (caar changes) posn)))))
+      (and changes (or (<= (caar changes) posn)
+                       (and (thunk-force atomic?)
+                            (<= (caar changes) sexp-end)))))))
 
-(defun el-search--changed-p (posn &optional revision)
+(defun el-search--changed-p (posn revision)
   ;; Non-nil when sexp after POSN contains a change
   (when (buffer-modified-p)
     (user-error "Buffer is modified - please save"))
@@ -255,34 +317,37 @@ Use variable `el-search--cached-changes' for caching."
       (while (and changes (<= (cdar changes) posn))
         (pop changes))
       (and changes
-           (< (caar changes) (scan-sexps posn 1))))))
-
-(defvar vc-git-diff-switches)
-(defun el-search--file-changed-p (file rev)
-  ;; FIXME: it would be better to calculate once a list of all changed
-  ;; files in the repository
-  (cl-callf file-truename file)
-  (when-let ((backend (vc-backend file)))
-    (ignore-errors
-      (let ((default-directory (file-name-directory file))
-            (vc-git-diff-switches nil)) ;FIXME: necessary e.g. for my init file -- why?
-        (and
-         (with-temp-buffer
-           (= 1 (vc-call-backend backend 'diff (list file) nil rev (current-buffer))))
-         (with-temp-buffer
-           (= 1 (vc-call-backend backend 'diff (list file) rev nil (current-buffer)))))))))
+           (< (caar changes) (el-search--end-of-sexp posn))))))
 
 (defun el-search-change--heuristic-matcher (&optional revision)
-  (lambda (file-name-or-buffer _)
-    (require 'vc)
-    (when-let ((file (if (stringp file-name-or-buffer)
-                         file-name-or-buffer
-                       (buffer-file-name file-name-or-buffer))))
-      (let ((default-directory (file-name-directory file)))
-        (el-search--file-changed-p
-         file
-         (funcall el-search-change-revision-transformer-function
-                  (or revision "HEAD") file))))))
+  (let* ((revision (or revision "HEAD"))
+         (get-changed-files-in-repo
+          (el-search-with-short-term-memory #'el-search--changed-files-in-repo))
+         (file-changed-p (el-search-with-short-term-memory
+                          (lambda (file-name-or-buffer)
+                            (require 'vc)
+                            (when-let ((file (if (stringp file-name-or-buffer)
+                                                 file-name-or-buffer
+                                               (buffer-file-name file-name-or-buffer))))
+                              (cl-callf el-search--file-truename-wstm file)
+                              (let ((default-directory (file-name-directory file)))
+                                (when-let ((backend (vc-backend file))
+                                           (root-dir
+                                            (condition-case err
+                                                (vc-call-backend backend 'root default-directory)
+                                              ;; Same handler as in `vc-root-dir'
+                                              (vc-not-supported
+                                               (unless (eq (cadr err) 'root)
+                                                 (signal (car err) (cdr err)))
+                                               nil))))
+                                  (cl-some
+                                   (apply-partially #'file-equal-p file)
+                                   (funcall get-changed-files-in-repo
+                                            root-dir
+                                            (funcall (or el-search-change-revision-transformer-function
+                                                         (lambda (rev _) rev))
+                                                     revision file))))))))))
+    (lambda (file-name-or-buffer _) (funcall file-changed-p file-name-or-buffer))))
 
 (el-search-defpattern change (&optional revision)
   "Matches the object if its text is part of a file change.
@@ -290,7 +355,10 @@ Use variable `el-search--cached-changes' for caching."
 Requires library \"diff-hl\".  REVISION defaults to the file's
 repository's HEAD commit and is a revision string.  Customize
 `el-search-change-revision-transformer-function' to control how
-REVISION is interpreted."
+REVISION is interpreted.
+
+This pattern-type does currently only work for git versioned
+files."
   (declare (heuristic-matcher #'el-search-change--heuristic-matcher))
   `(guard (el-search--change-p (point) ,(or revision "HEAD"))))
 
@@ -300,11 +368,50 @@ REVISION is interpreted."
 Requires library \"diff-hl\".  REVISION defaults to the file's
 repository's HEAD commit and is a revision string.  Customize
 `el-search-change-revision-transformer-function' to control how
-REVISION is interpreted."
+REVISION is interpreted.
+
+This pattern-type does currently only work for git versioned
+files."
   (declare (heuristic-matcher #'el-search-change--heuristic-matcher))
   `(guard (el-search--changed-p (point) ,(or revision "HEAD"))))
 
 
+;;;; `outermost' and `top-level'
+
+(el-search-defpattern outermost (pattern &optional not-pattern)
+    "Matches when PATTERN matches but the parent sexp does not.
+For toplevel expressions, this is equivalent to PATTERN.
+
+Optional NOT-PATTERN defaults to PATTERN; when given, match when
+PATTERN matches but the parent sexp is not matched by
+NOT-PATTERN.
+
+
+This pattern is useful to match only the outermost expression
+when subexpressions would match recursively.  For
+example, (outermost _) matches only top-level expressions.
+Another example: For the `change' pattern, any subexpression of a
+match is typically also an according change.  Wrapping the
+`change' pattern into `outermost' prevents el-search from
+descending into any found expression - only the outermost
+expression matching the `change' pattern will be matched."
+    `(and ,pattern
+          (not (guard (save-excursion
+                        (condition-case nil
+                            (progn
+                              (backward-up-list)
+                              (el-search--match-p
+                               ',(el-search-make-matcher (or not-pattern pattern))
+                               (save-excursion (el-search-read (current-buffer)))))
+                          (scan-error)))))))
+
+(el-search-defpattern top-level ()
+  "Matches any toplevel expression."
+  '(outermost _))
+
+
+;;; Sloppy pattern types for quick navigation
+
 ;;;; `keys'
 
 (defun el-search--match-key-sequence (keys expr)
@@ -340,39 +447,6 @@ matches any of these expressions:
   `(pred (el-search--match-key-sequence ,key-sequence)))
 
 
-;;;; `outermost' and `top-level'
-
-(el-search-defpattern outermost (pattern &optional not-pattern)
-    "Matches when PATTERN matches but the parent sexp does not.
-For toplevel expressions, this is equivalent to PATTERN.
-
-Optional NOT-PATTERN defaults to PATTERN; when given, match when
-PATTERN matches but the parent sexp is not matched by
-NOT-PATTERN.
-
-
-This pattern is useful to match only the outermost expression
-when subexpressions would match recursively.  For
-example, (outermost _) matches only top-level expressions.
-Another example: For the `change' pattern, any subexpression of a
-match is typically also an according change.  Wrapping the
-`change' pattern into `outermost' prevents el-search from
-descending into any found expression - only the outermost
-expression matching the `change' pattern will be matched."
-    `(and ,pattern
-          (not (guard (save-excursion
-                        (condition-case nil
-                            (progn
-                              (backward-up-list)
-                              (el-search--match-p
-                               ',(el-search--matcher (or not-pattern pattern))
-                               (save-excursion (read (current-buffer)))))
-                          (scan-error)))))))
-
-(el-search-defpattern top-level ()
-  "Matches any toplevel expression."
-  '(outermost _))
-
 
 ;;; Patterns for stylistic rewriting and syntactical simplification
 
